import torch
from torch import nn
import torch.nn.functional as F
import pickle
import math
import arrow
import numpy as np


class AriadnePredictData:
    @staticmethod
    def buildNodeFeatures(nodes_dict, node_id):
        attr_id = nodes_dict[node_id]

        pos_d = np.array(attr_id["pos_d"])
        pos_d = (pos_d + 1) / 2

        dir = np.array(attr_id["pos_d"])
        angle = math.atan(dir[1] / dir[0])
        angle = (angle + math.pi * 0.5) / math.pi

        vector = [
            float(pos_d[0]),
            float(pos_d[1]),
            float(angle),
            float(attr_id["radius"]),
            float(attr_id["color"][0]),
            float(attr_id["color"][1]),
            float(attr_id["color"][2]),
        ]
        return vector

    @staticmethod
    def getAllPairsFromMemory(path):
        with open(path, "rb") as handle:
            graph = pickle.load(handle)

        nodes = graph["nodes"]
        pred_edges = graph["pred_edges"]

        return AriadnePredictData.makeTensors(nodes, pred_edges)

    @staticmethod
    def makeTensors(nodes, pred_edges):
        rv = []
        for e in pred_edges:
            x0 = AriadnePredictData.buildNodeFeatures(nodes, node_id=e[0])
            x0 = torch.Tensor(x0)

            x1 = AriadnePredictData.buildNodeFeatures(nodes, node_id=e[1])
            x1 = torch.Tensor(x1)

            rv.append({"node_0": e[0], "node_1": e[1], "feat_0": x0, "feat_1": x1})
        return rv

    @staticmethod
    def getAllPairs(nodes, pred_edges):
        return AriadnePredictData.makeTensors(nodes, pred_edges)


class SiameseNetwork(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim):
        super(SiameseNetwork, self).__init__()

        self.linear1 = nn.Linear(in_features=input_dim, out_features=hidden_dim)
        self.bn1 = nn.BatchNorm1d(num_features=hidden_dim)

        self.linear2 = nn.Linear(in_features=hidden_dim, out_features=hidden_dim)
        self.bn2 = nn.BatchNorm1d(num_features=hidden_dim)

        self.linear3 = nn.Linear(in_features=hidden_dim, out_features=output_dim)

    def forward_once(self, x):
        y = F.relu(self.bn1(self.linear1(x)))
        y = F.relu(self.bn2(self.linear2(y)))
        return F.normalize(self.linear3(y))

    def forward(self, input_t, input_p, input_n):
        output_t = self.forward_once(input_t)
        output_p = self.forward_once(input_p)
        output_n = self.forward_once(input_n)

        return output_t, output_p, output_n


class NN(object):
    def __init__(self, device, checkpoint_path):
        self.device = device
        self.checkpoint_path = checkpoint_path
        self.model = SiameseNetwork(input_dim=7, hidden_dim=32, output_dim=16)

        # WEIGHTS LOADING
        self.model.load_state_dict(torch.load(self.checkpoint_path))
        self.model.eval()

    def gaussian(self, val):
        return np.e ** (-(val**2))

    def predict(self, data, threshold=0.0):
        t0 = arrow.utcnow()

        out = []
        for sample in data:
            t1 = arrow.utcnow()

            z0 = self.model.forward_once(sample["feat_0"].unsqueeze(0))
            z1 = self.model.forward_once(sample["feat_1"].unsqueeze(0))
            prob = torch.exp(-torch.cdist(z0, z1) ** 2).squeeze().detach().cpu().numpy()

            # logging
            list_nodes = [sample["node_0"], sample["node_1"]]
            tot_time = (arrow.utcnow() - t1).total_seconds() * 1000
            # print(f"Nodes: {list_nodes}\t| Score: {prob:.4f} \t| Time: {tot_time:.4f} ms")

            if prob > threshold:
                out.append({"node_0": list_nodes[0], "node_1": list_nodes[1], "score": prob})

        # cprint(f"time NN prediction: {(arrow.utcnow()-t0).total_seconds() * 1000:.4f} ms", "yellow")
        return out

    def predictBatch(self, data, threshold=0.5, log=False):
        if len(data) == 0:
            return []

        t0 = arrow.utcnow()
        loader = torch.utils.data.DataLoader(data, batch_size=len(data), shuffle=False, num_workers=0)

        out = []
        for batch in loader:
            z0 = self.model.forward_once(batch["feat_0"])
            z1 = self.model.forward_once(batch["feat_1"])

            for it, (zz0, zz1) in enumerate(zip(z0, z1)):
                prob = (
                    torch.exp(-torch.cdist(zz0.unsqueeze(0), zz1.unsqueeze(0)) ** 2).squeeze().detach().cpu().numpy()
                )

                if prob > threshold:
                    list_nodes = [int(batch["node_0"][it].detach().numpy()), int(batch["node_1"][it].detach().numpy())]
                    if log:
                        print(f"Nodes: {list_nodes}\t| Score: {prob:.4f}")

                    out.append({"node_0": list_nodes[0], "node_1": list_nodes[1], "score": prob})

        if log:
            print(f"Tot Time: {(arrow.utcnow()-t0).total_seconds() * 1000:.4f} ms")
        return out
